# Copyright 2021 VMware, Inc.
# SPDX-License-Identifier: Apache-2.0
from typing import List

from vdk.internal.builtin_plugins.config.job_config import JobConfigKeys
from vdk.internal.core.config import Configuration
from vdk.internal.core.config import ConfigurationBuilder

NOTIFICATION_ENABLED = "NOTIFICATION_ENABLED"
NOTIFICATION_JOB_LOG_URL_PATTERN = "NOTIFICATION_JOB_LOG_URL_PATTERN"
NOTIFICATION_EMAIL_CC_LIST = "NOTIFICATION_EMAIL_CC_LIST"
NOTIFICATION_SENDER = "NOTIFICATION_SENDER"
NOTIFICATION_SMTP_HOST = "NOTIFICATION_SMTP_HOST"
NOTIFICATION_SMTP_PORT = "NOTIFICATION_SMTP_PORT"
NOTIFICATION_SMTP_USE_TLS = "NOTIFICATION_SMTP_USE_TLS"
NOTIFICATION_SMTP_LOGIN_USERNAME = "NOTIFICATION_SMTP_LOGIN_USERNAME"
NOTIFICATION_SMTP_LOGIN_PASSWORD = "NOTIFICATION_SMTP_LOGIN_PASSWORD"
NOTIFICATION_SMTP_DEBUG_LEVEL = "NOTIFICATION_SMTP_DEBUG_LEVEL"


class NotificationConfiguration:
    def __init__(self, config: Configuration):
        self.__config = config

    def get_job_log_url_pattern(self) -> str:
        return str(self.__config[NOTIFICATION_JOB_LOG_URL_PATTERN])

    def get_email_cc_list(self) -> List[str]:
        # TODO: we should considure ConfigurationBuilder to support List type for configuration values
        # and automatically parses them as below
        cc_list = str(self.__config[NOTIFICATION_EMAIL_CC_LIST])
        cc_result = cc_list.split(";") if cc_list else []
        cc_result = [cc.strip() for cc in cc_result if len(cc.strip()) > 0]
        return cc_result

    def get_sender(self) -> str:
        return str(self.__config[NOTIFICATION_SENDER])

    def get_notification_enabled(self):
        return bool(self.__config[NOTIFICATION_ENABLED])


class SmtpConfiguration:
    def __init__(self, config: Configuration):
        self.__config = config

    def get_host(self) -> str:
        return str(self.__config[NOTIFICATION_SMTP_HOST])

    def get_port(self) -> int:
        return int(self.__config[NOTIFICATION_SMTP_PORT])

    def get_use_tls(self) -> bool:
        return bool(self.__config[NOTIFICATION_SMTP_USE_TLS])

    def get_login_username(self) -> str:
        return str(self.__config[NOTIFICATION_SMTP_LOGIN_USERNAME])

    def get_login_password(self) -> str:
        return str(self.__config[NOTIFICATION_SMTP_LOGIN_PASSWORD])

    def get_debug_level(self) -> int:
        return int(self.__config[NOTIFICATION_SMTP_DEBUG_LEVEL])


def __add_job_notified_configuration_definitions(config_builder: ConfigurationBuilder):
    config_builder.add(
        key=JobConfigKeys.NOTIFIED_ON_JOB_FAILURE_USER_ERROR.value,
        default_value="",
        description="Semicolon-separated list of email addresses to be notified on job execution "
        "failure caused by user code or user configuration problem. "
        "For example: if the job contains an SQL script with syntax error.",
    )
    config_builder.add(
        key=JobConfigKeys.NOTIFIED_ON_JOB_FAILURE_PLATFORM_ERROR.value,
        default_value="",
        description="Semicolon-separated list of email addresses to be notified on job execution failure "
        "caused by a platform problem, including job execution delays.",
    )
    config_builder.add(
        key=JobConfigKeys.NOTIFIED_ON_JOB_SUCCESS.value,
        default_value="",
        description="Semicolon-separated list of email addresses to be notified on job execution success.",
    )
    config_builder.add(
        key=JobConfigKeys.NOTIFIED_ON_JOB_DEPLOY.value,
        default_value="",
        description="Semicolon-separated list of email addresses to be notified of job deployment outcome",
    )
    config_builder.add(
        key=JobConfigKeys.ENABLE_ATTEMPT_NOTIFICATIONS.value,
        default_value=False,
        description="Flag to enable or disable the email notifications sent to the recipients listed above "
        "for each data job run attempt.",
    )


def __add_smtp_configuration_definitions(config_builder):
    config_builder.add(
        key=NOTIFICATION_SMTP_HOST,
        default_value="smtp.vmware.com",
        description="The SMTP host used for sending the notification email.",
    )
    config_builder.add(
        key=NOTIFICATION_SMTP_PORT,
        default_value=25,
        description="The SMTP port used for to send the notification email.",
    )
    config_builder.add(
        key=NOTIFICATION_SMTP_USE_TLS,
        default_value=False,
        description="Set to true if you want to use secure connection over TLS/SSL.",
    )
    config_builder.add(
        key=NOTIFICATION_SMTP_LOGIN_USERNAME,
        default_value="",
        description="Login username to use when connecting to SMTP server. If empty no login is attempted.",
    )
    config_builder.add(
        key=NOTIFICATION_SMTP_LOGIN_PASSWORD,
        default_value="",
        description="Login password to use when connecting to SMTP server.",
    )
    config_builder.add(
        key=NOTIFICATION_SMTP_DEBUG_LEVEL,
        default_value=1,
        description="SMTP debug level. A non-zero value results in debug messages for connection "
        "and for all messages sent to and received from the SMTP server.",
    )


def __add_job_notified_configuration_definitions(config_builder: ConfigurationBuilder):
    config_builder.add(
        key=JobConfigKeys.NOTIFIED_ON_JOB_FAILURE_USER_ERROR.value,
        default_value="",
        description="Semicolon-separated list of email addresses to be notified on job execution "
        "failure caused by user code or user configuration problem. "
        "For example: if the job contains an SQL script with syntax error.",
    )
    config_builder.add(
        key=JobConfigKeys.NOTIFIED_ON_JOB_FAILURE_PLATFORM_ERROR.value,
        default_value="",
        description="Semicolon-separated list of email addresses to be notified on job execution failure."
        "caused by a platform problem, including job execution delays.",
    )
    config_builder.add(
        key=JobConfigKeys.NOTIFIED_ON_JOB_SUCCESS.value,
        default_value="",
        description="Semicolon-separated list of email addresses to be notified on job execution success.",
    )
    config_builder.add(
        key=JobConfigKeys.NOTIFIED_ON_JOB_DEPLOY.value,
        default_value="",
        description="Semicolon-separated list of email addresses to be notified of job deployment outcome.",
    )
    config_builder.add(
        key=JobConfigKeys.ENABLE_ATTEMPT_NOTIFICATIONS.value,
        default_value=False,
        description="Flag to enable or disable the email notifications sent to the recipients listed above "
        "for each data job run attempt.",
    )


def add_definitions(config_builder: ConfigurationBuilder):
    config_builder.add(
        key=NOTIFICATION_ENABLED,
        default_value=False,
        description="Set to true if you want to enable notifications regardless if it is run locally or not.",
    )
    config_builder.add(
        key=NOTIFICATION_JOB_LOG_URL_PATTERN,
        default_value=(
            "https://example-job-log-url.com/{op_id},{start_time_ms},{end_time_ms}"
        ),
        description=(
            "The URL template used to find the full log of a particular job. "
            "It is further parametrized with the job's op_id and start and end times "
            "by replacing the bracketed parts of the url with the respective values. "
            "(For example, '{op_id}' becomes the actual op_id of the job.)"
        ),
    )
    config_builder.add(
        key=NOTIFICATION_EMAIL_CC_LIST,
        default_value="",
        description="A semicolon-separated separated string of email addresses to be CC'd in the notification email.",
    )
    config_builder.add(
        key=NOTIFICATION_SENDER,
        default_value="data-pipelines@vmware.com",
        description="The email address, from which notification emails are sent.",
    )

    __add_smtp_configuration_definitions(config_builder)
    __add_job_notified_configuration_definitions(config_builder)



# Copyright 2021 VMware, Inc.
# SPDX-License-Identifier: Apache-2.0
import logging
import pathlib
import textwrap
from typing import List
from typing import Optional

from vdk.api.job_input import IJobArguments
from vdk.api.job_input import IJobInput
from vdk.api.job_input import ITemplate
from vdk.internal.builtin_plugins.connection.impl.router import ManagedConnectionRouter
from vdk.internal.builtin_plugins.connection.managed_connection_base import (
    ManagedConnectionBase,
)
from vdk.internal.builtin_plugins.ingestion.ingester_router import IngesterRouter
from vdk.internal.builtin_plugins.job_properties.properties_router import (
    PropertiesRouter,
)
from vdk.internal.builtin_plugins.job_properties.Propertiesnotavailable import (
    PropertiesNotAvailable,
)
from vdk.internal.builtin_plugins.run.execution_results import ExecutionResult
from vdk.internal.builtin_plugins.run.execution_state import ExecutionStateStoreKeys
from vdk.internal.builtin_plugins.run.sql_argument_substitutor import (
    SqlArgumentSubstitutor,
)
from vdk.internal.core import errors
from vdk.internal.core.errors import UserCodeError
from vdk.internal.core.statestore import CommonStoreKeys
from vdk.internal.core.statestore import StateStore


# TODO make this extensible so new job inputs can be added by plugins
class JobInput(IJobInput):
    """
    Implementation of JobInput interface.
    It just forwards to actual implementation of a given functionality and should do nothing else.
    """

    def __init__(
        self,
        managed_connection_builder: ManagedConnectionRouter,
        statestore: StateStore,
        properties_router: PropertiesRouter,
        job_arguments: IJobArguments,
        templates: ITemplate,
        ingester: IngesterRouter,
    ):
        """Constructor."""

        self.__managed_connection_builder = managed_connection_builder
        self.__statestore = statestore
        self.__properties_router = properties_router
        self.__vdk_internal_telemetry = None
        self.__ingester = ingester
        self.__job_arguments = job_arguments
        self.__templates = templates

    # Connections

    def get_managed_connection(self) -> ManagedConnectionBase:
        return self.__managed_connection_builder.open_default_connection()

    def get_arguments(self) -> dict:
        return self.__job_arguments.get_arguments()

    # Properties

    def get_property(self, name, default_value=None):
        return self.__properties_router.get_properties_impl().get_property(
            name, default_value
        )

    def get_all_properties(self):
        return self.__properties_router.get_properties_impl().get_all_properties()

    def set_all_properties(self, properties):
        return self.__properties_router.get_properties_impl().set_all_properties(
            properties
        )

    def _substitute_query_params(self, sql: str):
        sql = textwrap.dedent(sql).strip("\n") + "\n"
        query = sql
        if isinstance(
            self.__properties_router.get_properties_impl(), PropertiesNotAvailable
        ):
            logging.getLogger(__name__).info(
                "Data Job Properties has not been initialized., "
                "so I won't be able to provide query properties substitution capabilities from job properties."
                "If passed job arguments will still be used"
            )
        else:
            sql_args = self.get_all_properties()
            sql_args.update(self.get_execution_properties())
            query = SqlArgumentSubstitutor(sql_args).substitute(query)

        sql_args = self.get_arguments()
        if not sql_args or type(sql_args) != dict:
            logging.getLogger(__name__).debug(
                "No arguments are passed for Data Job, "
                "so I won't be able to provide query parameter substitution capabilities with job arguments."
            )
        else:
            query = SqlArgumentSubstitutor(sql_args).substitute(query)

        return query

    def execute_query(self, sql: str):
        if not sql or not sql.strip():
            raise UserCodeError("Trying to execute an empty SQL query.")

        job_name = self.__statestore.get(ExecutionStateStoreKeys.JOB_NAME)
        op_id = self.__statestore.get(CommonStoreKeys.OP_ID)

        query = self._substitute_query_params(sql)
        query = "\n".join(
            ["-- job_name: {job_name}", "-- op_id: {op_id}", "{query}"]
        ).format(job_name=job_name, op_id=op_id, query=query)

        connection = self.get_managed_connection()
        return connection.execute_query(query)

    def send_object_for_ingestion(
        self,
        payload: dict,
        destination_table: Optional[str] = None,
        method: Optional[str] = None,
        target: Optional[str] = None,
        collection_id=None,
    ):
        self.__ingester.send_object_for_ingestion(
            payload,
            destination_table,
            method,
            target,
            collection_id,
        )

    def send_tabular_data_for_ingestion(
        self,
        rows: iter,
        column_names: List,
        destination_table: Optional[str] = None,
        method: Optional[str] = None,
        target: Optional[str] = None,
        collection_id: Optional[str] = None,
    ):
        self.__ingester.send_tabular_data_for_ingestion(
            rows,
            column_names,
            destination_table,
            method,
            target,
            collection_id,
        )

    def execute_template(
        self, template_name: str, template_args: dict
    ) -> ExecutionResult:
        if self.__templates:
            result = self.__templates.execute_template(template_name, template_args)

            return result
        else:
            raise NotImplemented("Templates not wired to JobInput")

    def get_name(self) -> str:
        pass

    def get_job_directory(self) -> pathlib.Path:
        pass

    def get_execution_properties(self) -> dict:
        start_time = self.__statestore.get(CommonStoreKeys.START_TIME)
        return {
            "pa__execution_id": self.__statestore.get(CommonStoreKeys.EXECUTION_ID),
            "pa__job_start_unixtime": str(int(start_time.timestamp())),
            "pa__job_start_ts_expr": f"cast ({start_time.timestamp()} as timestamp)",
            "pa__op_id": self.__statestore.get(CommonStoreKeys.OP_ID),
        }



# Copyright 2021 VMware, Inc.
# SPDX-License-Identifier: Apache-2.0
import logging
from abc import abstractmethod
from types import TracebackType
from typing import Any
from typing import cast
from typing import List
from typing import Optional
from typing import Type

from tenacity import before_sleep_log
from tenacity import retry
from tenacity import retry_if_exception_type
from tenacity import stop_after_attempt
from tenacity import wait_exponential
from vdk.api.job_input import IManagedConnection
from vdk.internal.builtin_plugins.connection.managed_cursor import ManagedCursor
from vdk.internal.builtin_plugins.connection.pep249.interfaces import PEP249Connection
from vdk.internal.core import errors
from vdk.internal.util.decorators import closing_noexcept_on_close

logger = logging.getLogger(__name__)


class ManagedConnectionBase(PEP249Connection, IManagedConnection):
    """
    Different database providers can subclass this class to provide raw connection (by implement _connect)

    The raw connection will be managed by the application.
    Optionally further customization is allowed by overriding _cursor_execute and_before_cursor_execute
    """

    def __init__(
        self, log: logging.Logger = logger, db_con: Optional[PEP249Connection] = None
    ):
        """
        this constructor MUST be called by inheritors
        """
        if log:
            self._log = log
        else:
            self._log = logging.getLogger(__name__)
        self._is_db_con_open: bool = db_con is not None
        self._db_con: Optional[PEP249Connection] = db_con

    @abstractmethod
    def _connect(self) -> PEP249Connection:
        """
        this method MUST be implemented by inheritors and should return PEP249 Connection object (unmanaged)
        """
        raise NotImplementedError

    # Retry to connect on exception and backoff exponentially in
    # 30s, 1m, 2m, 4m
    @retry(
        stop=stop_after_attempt(5),
        wait=wait_exponential(multiplier=30, min=30, max=240),
        retry=retry_if_exception_type(Exception),
        before_sleep=before_sleep_log(logger, logging.DEBUG),
        reraise=True,
    )
    def connect(self) -> PEP249Connection:
        """
        :return: PEP249 Connection object (unmanaged)
        """
        if not self._is_db_con_open:
            db_con = self._connect()
            self._log.debug("Established %s", str(db_con))
            self._is_db_con_open = True
            self._db_con = db_con
        return self._db_con

    def cursor(self, *args, **kwargs):
        if hasattr(self._db_con, "cursor"):
            return ManagedCursor(self._db_con.cursor(*args, **kwargs), self._log)
        return super().cursor()

    def commit(self, *args, **kwargs):
        if hasattr(self._db_con, "commit"):
            return self._db_con.commit(*args, **kwargs)
        return super().commit()

    def rollback(self, *args, **kwargs):
        if hasattr(self._db_con, "rollback"):
            return self._db_con.rollback(*args, **kwargs)
        return super().rollback()

    def close(self, *args, **kwargs) -> None:
        """
        Close database connection. No op if it is already closed.
        """
        try:
            if self._is_db_con_open:
                self._log.debug(f"Closing database connection {self} ... ")
                self._is_db_con_open = False
                self._db_con.close()
                self._log.debug("Closing database connection SUCCEEDED.")
        except Exception as e:
            self._log.exception(
                "Closing database connection FAILED. No problem, I'm continuing as if nothing happened.",
                e,
            )

    # @abstractmethod # inherit optionally
    def _before_cursor_execute(self, cur: ManagedCursor) -> None:
        pass

    # @abstractmethod # inherit optionally
    def _cursor_execute(self, cur: ManagedCursor, query: str) -> None:
        cur.execute(query)  # no need to log around this method.

    def execute_query(self, query: str) -> List[List[Any]]:
        """
        Execute SQL query.
        """
        # TODO: configurable generic re-try.
        try:
            #  query = query.encode('utf-8') #TODO make it a validation step and fail if query is not UTF-8
            #  query = query.decode('utf-8') #TODO make it a validation step and fail if query is not UTF-8
            with closing_noexcept_on_close(self.__cursor()) as cur:
                self._before_cursor_execute(cur)
                self._cursor_execute(cur, query)

                try:
                    """
                    1. According to PEP 249 fetchall() should throw an exception when there is no result set
                    produced by cursor.execute() (say insert into table).
                    2. In pyodbc the cursor.rowcount property is always -1 => can rely upon.
                    3. In impyla there is a handy property cursor.has_result_set.
                    But it is not in PEP 249 and not supported by pyodbc implementation
                    4. The only solution found so far is to try/catch the fetchall() call
                    and swallow the exception in very narrow set of cases.
                    """
                    # TODO support for fetchmany
                    res = cur.fetchall()
                except Exception as pe:
                    res = None
                    # TODO: this will likely not work in other dbs.
                    if str(pe) in (
                        "No results.  Previous SQL was not a query.",  # message in pyodbc
                        "Trying to fetch results on an operation with no results.",  # message in impyla
                        "no results to fetch",  # psycopg: ProgrammingError: no results to fetch
                    ):
                        self._log.debug(
                            "Fetching all results from query SUCCEEDED. Query does not produce results (e.g. DROP "
                            "TABLE). "
                        )
                    else:
                        self._log.debug("Fetching all results from query FAILED.")
                        raise
                    # we return None in case of DML. This is not PEP249 compliant, but is more convenient
                return cast(List[List[Any]], res)
        except Exception as e:
            # without knowing anything most likely issue in SQL query is due to job owner.
            # Plugins may override this.
            blamee = errors.ResolvableBy.USER_ERROR
            errors.log_and_rethrow(
                blamee,
                self._log,
                what_happened="Executing query FAILED.",
                why_it_happened=errors.MSG_WHY_FROM_EXCEPTION(e),
                consequences=errors.MSG_CONSEQUENCE_DELEGATING_TO_CALLER__LIKELY_EXECUTION_FAILURE,
                countermeasures=errors.MSG_COUNTERMEASURE_FIX_PARENT_EXCEPTION,
                exception=e,
            )
            return []

    def __cursor(self) -> ManagedCursor:
        conn = self.connect()
        cur = conn.cursor()
        return ManagedCursor(cur, self._log)

    # @abstractmethod # inherit optionally e.g. in case database does not support select 1 for checks
    def _is_connected(self) -> bool:
        if None is self._is_db_con_open:
            return False
        if False is self._is_db_con_open:
            return False
        try:
            self.__cursor().execute(" select 1 -- Testing if connection is alive.")
            return True
        except Exception as e:
            self._log.debug(
                f"Connection {self} is disconnected ('select 1' returned {e})"
            )
        return False

    def get_managed_connection(self) -> PEP249Connection:
        return self

    def __enter__(self) -> Any:
        return self

    def __exit__(
        self,
        exception_type: Optional[Type[BaseException]],
        exception_value: Optional[BaseException],
        traceback: Optional[TracebackType],
    ) -> None:  # @UnusedVariable
        self.close()

    def __del__(self) -> None:
        self.close()

    def __str__(self) -> str:
        return "ManagedConnection[ isConnected:{} {} ]".format(
            self._is_db_con_open, self._db_con
        )



# Copyright 2021 VMware, Inc.
# SPDX-License-Identifier: Apache-2.0
import logging
from enum import Enum

import click
from vdk.api.plugin.hook_markers import hookimpl
from vdk.internal.builtin_plugins.version import version
from vdk.internal.builtin_plugins.version.new_version_check import Package
from vdk.internal.core.config import ConfigurationBuilder
from vdk.internal.core.context import CoreContext

log = logging.getLogger(__name__)


class ConfigKey(str, Enum):
    PACKAGE_NAME = "PACKAGE_NAME"
    PACKAGE_INDEX = "PACKAGE_INDEX"
    VERSION_CHECK_PLUGINS = "VERSION_CHECK_PLUGINS"


def new_package(package_index, package_name):
    return Package(package_name, package_index)


class NewVersionCheckPlugin:
    @hookimpl(tryfirst=True)
    def vdk_configure(self, config_builder: ConfigurationBuilder) -> None:
        config_builder.add(
            key=ConfigKey.PACKAGE_NAME,
            default_value="vdk-core",
            description="Set distribution package name of the library to check for new version",
        )
        config_builder.add(
            key=ConfigKey.PACKAGE_INDEX,
            default_value="https://pypi.org",
            description="Set distribution package name of the library to check for new version",
        )
        config_builder.add(
            key=ConfigKey.VERSION_CHECK_PLUGINS,
            default_value=True,
            description="Set to true if plugins should be checked for new version otherwise false",
        )

    @hookimpl
    def vdk_initialize(self, context: CoreContext) -> None:
        try:
            cfg = context.configuration
            package_name = cfg.get_value(ConfigKey.PACKAGE_NAME)
            package_index = cfg.get_value(ConfigKey.PACKAGE_INDEX)
            self.check_version(package_name, package_index)

            if cfg.get_value(ConfigKey.VERSION_CHECK_PLUGINS):
                log.debug("Will check for newer versions all installed plugins.")
                for dist_name, _ in version.list_installed_plugins():
                    self.check_version(dist_name, package_index, is_plugin=True)
        except Exception as e:
            log.info(
                f"Could not check for new version release. "
                f"Error was {e}. We are ignoring the error."
            )

    @staticmethod
    def check_version(package_name, package_index, is_plugin=False):
        if new_package(package_index, package_name).check():
            # if package index is not specified that we fetch it from pip repo (pypi.org)
            extra_index_if_needed = (
                f"--extra-index-url {package_index}" if package_index else ""
            )
            message = (
                f"New version for a plugin {package_name} is Available"
                if is_plugin
                else f"New Version of {package_name} is Available"
            )
            # We are using eager strategy so that if there's new version of a plugin that satisfies the requirements
            # of the main package we want to upgrade to it.
            pip_command = "pip install --upgrade-strategy eager -U"
            click.echo(
                f"""
                **********************************************************************************************************************

                                                {message}

                  Please update to latest version by using:
                  {pip_command} {package_name} {extra_index_if_needed}

                **********************************************************************************************************************
                """,
                err=True,
            )



# Copyright 2021 VMware, Inc.
# SPDX-License-Identifier: Apache-2.0
import logging


class PEP249Connection:
    """
    Interface and protocol for PEP 249 compatible connection
    """

    def close(self):
        """PEP 249
        Close the connection now (rather than whenever .__del__() is called).
        The connection will be unusable from this point forward; an Error (or subclass) exception will be raised
        if any operation is attempted with the connection. The same applies to all cursor objects trying to use
        the connection. Note that closing a connection without committing the changes first will cause an
        implicit rollback to be performed.
        """
        raise NotImplementedError

    def commit(self):
        # PEP 249
        raise NotImplementedError

    def rollback(self):
        # PEP 249
        raise NotImplementedError

    def cursor(self):
        """PEP 249
        Return a new Cursor Object using the connection.
        If the database does not provide a direct cursor concept, the module will have to emulate cursors using
        other means to the extent needed by PEP 249 specification.
        """
        raise NotImplementedError


class PEP249Cursor:
    """Abstract representation of Cursor"""

    def __init__(self, cursor, log):
        if not log:
            log = logging.getLogger(__name__)
        self._cur = cursor
        self._log = log

    @property
    def description(self):
        return self._cur.description

    @property
    def rowcount(self):
        return self._cur.rowcount

    def callproc(self, procname, parameters=None):
        return self._cur.callproc(procname, parameters)

    def close(self):
        return self._cur.close()

    def execute(self, operation, parameters=None):
        return self._cur.execute(operation, parameters)

    def executemany(self, operation, seq_of_parameters):
        return self._cur.executemany(operation, seq_of_parameters)

    def fetchone(self):
        return self._cur.fetchone()

    def fetchmany(self, size=None):
        return self._cur.fetchmany(size)

    def fetchall(self):
        return self._cur.fetchall()

    def nextset(self):
        return self._cur.nextset()

    def _get_arraysize(self):
        return self._cur._get_arraysize()

    def _set_arraysize(self, arraysize):
        return self._cur._set_arraysize()

    arraysize = property(_get_arraysize, _set_arraysize)

    def setinputsizes(self, sizes):
        return self._cur.setinputsizes(sizes)

    def setoutputsize(self, size, column=None):
        return self._cur.setoutputsize(size, column)

    def __iter__(self):
        return self._cur.__iter__()

    def __next__(self):
        return self._cur.__next__()

    def next(self):
        # for py2 compat
        return self.__next__()



# Copyright 2021 VMware, Inc.
# SPDX-License-Identifier: Apache-2.0
import logging
from typing import Callable
from typing import Dict
from typing import Union

from vdk.api.plugin.plugin_input import IManagedConnectionRegistry
from vdk.internal.builtin_plugins.config.vdk_config import DB_DEFAULT_TYPE
from vdk.internal.builtin_plugins.connection.impl.wrapped_connection import (
    WrappedConnection,
)
from vdk.internal.builtin_plugins.connection.managed_connection_base import (
    ManagedConnectionBase,
)
from vdk.internal.builtin_plugins.connection.pep249.interfaces import PEP249Connection
from vdk.internal.core import errors
from vdk.internal.core.config import Configuration


class ManagedConnectionRouter(IManagedConnectionRegistry):
    """
    Create ManagedConnection by routing the configured database plugin.
    Configuration is controlled by DB_DEFAULT_TYPE for defualt connection.
    Or specfific connection can be specified by open_connection(dbtype)
    In both cases dbtype must match the string in which the plugin register itself with.
    """

    def __init__(self, cfg: Configuration):
        self._cfg: Configuration = cfg
        self._log: logging.Logger = logging.getLogger(__name__)
        self._connections: Dict[str, ManagedConnectionBase] = dict()
        self._connection_builders: Dict[
            str, Callable[[], ManagedConnectionBase]
        ] = dict()

    def add_open_connection_factory_method(
        self,
        dbtype: str,
        open_connection_func: Callable[
            [], Union[ManagedConnectionBase, PEP249Connection]
        ],
    ) -> None:
        """
        Add new connection factory method. See parent doc for more.
        """
        self._connection_builders[dbtype] = open_connection_func

    def open_default_connection(self) -> ManagedConnectionBase:
        """
        Open connection to the database configured as default (by db_default_type option).
        """
        dbtype = self._cfg.get_value(DB_DEFAULT_TYPE)
        if dbtype is None:
            if len(self._connection_builders) == 1:
                dbtype = list(self._connection_builders.keys())[0]
                self._log.debug(
                    f"DB_DEFAULT_TYPE has not been set. Found one database supported: {dbtype}. Will use it as default"
                )

        return self.open_connection(dbtype)

    def open_connection(self, dbtype: str) -> ManagedConnectionBase:
        """
        :param dbtype: The type of connection to open. It needs to have been registered before that by add_connection_builder
        or it will thrown an error
        :return: the new connection if succesfull or throws an expception
        """
        if dbtype in self._connections:
            conn = self._connections[dbtype]
            if conn._is_connected():
                return conn
            else:
                conn._connect()
                return conn
        self._log.debug(f"Connection to {dbtype} is missing. Will try to connect")
        if dbtype in self._connection_builders:
            return self.__create_connection(dbtype)
        errors.log_and_throw(
            to_be_fixed_by=errors.ResolvableBy.CONFIG_ERROR,
            log=self._log,
            what_happened=f"Provided configuration variable for {DB_DEFAULT_TYPE} has invalid value.",
            why_it_happened=f"VDK was run with {DB_DEFAULT_TYPE}={dbtype}, however {dbtype} is invalid value for this variable.",
            consequences=errors.MSG_CONSEQUENCE_DELEGATING_TO_CALLER__LIKELY_EXECUTION_FAILURE,
            countermeasures=f"Provide either valid value for {DB_DEFAULT_TYPE} or install database plugin that supports this type. "
            f"Currently possible values are {list(self._connection_builders.keys())}",
        )

    def __create_connection(self, dbtype):
        conn = self._connection_builders[dbtype]()
        if isinstance(conn, ManagedConnectionBase):
            self._connections[dbtype] = conn
        elif conn is None:
            errors.log_and_throw(
                to_be_fixed_by=errors.ResolvableBy.CONFIG_ERROR,
                log=self._log,
                what_happened=f"Could not create new connection of db type {dbtype}.",
                why_it_happened=f"VDK was run with {DB_DEFAULT_TYPE}={dbtype}, however no valid connection was created.",
                consequences=errors.MSG_CONSEQUENCE_DELEGATING_TO_CALLER__LIKELY_EXECUTION_FAILURE,
                countermeasures=f"Seems to be a bug in the plugin for dbtype {dbtype}. Make sure it's correctly installed. "
                f"If upgraded recently consider reverting to previous version. Or use another db type. "
                f"Currently possible values are {list(self._connection_builders.keys())}",
            )
        else:
            log = logging.getLogger(conn.__class__.__name__)
            conn.close()  # we will let ManagedConnection to open it when needed.
            self._connections[dbtype] = WrappedConnection(
                log, self._connection_builders[dbtype]
            )
        return self._connections[dbtype]



# Copyright 2021 VMware, Inc.
# SPDX-License-Identifier: Apache-2.0
from vdk.internal.core.config import Configuration
from vdk.internal.core.config import ConfigurationBuilder

INGESTER_NUMBER_OF_WORKER_THREADS = "INGESTER_NUMBER_OF_WORKER_THREADS"
INGESTER_PAYLOAD_SIZE_BYTES_THRESHOLD = "INGESTER_PAYLOAD_SIZE_BYTES_THRESHOLD"
INGESTER_OBJECTS_QUEUE_SIZE = "INGESTER_OBJECTS_QUEUE_SIZE"
INGESTER_PAYLOADS_QUEUE_SIZE = "INGESTER_PAYLOADS_QUEUE_SIZE"
INGESTER_LOG_UPLOAD_ERRORS = "INGESTER_LOG_UPLOAD_ERRORS"
INGESTION_PAYLOAD_AGGREGATOR_TIMEOUT_SECONDS = (
    "INGESTION_PAYLOAD_AGGREGATOR_TIMEOUT_SECONDS"
)
INGESTER_SHOULD_RAISE_EXCEPTION_ON_FAILURE = (
    "INGESTER_SHOULD_RAISE_EXCEPTION_ON_FAILURE"
)


class IngesterConfiguration:
    def __init__(self, config: Configuration):
        self.__config = config

    def get_number_of_worker_threads(self) -> int:
        return int(self.__config.get_value(INGESTER_NUMBER_OF_WORKER_THREADS))

    def get_payload_size_bytes_threshold(self) -> int:
        return int(self.__config.get_value(INGESTER_PAYLOAD_SIZE_BYTES_THRESHOLD))

    def get_objects_queue_size(self) -> int:
        return int(self.__config.get_value(INGESTER_OBJECTS_QUEUE_SIZE))

    def get_payloads_queue_size(self) -> int:
        return int(self.__config.get_value(INGESTER_PAYLOADS_QUEUE_SIZE))

    def get_should_log_upload_errors(self) -> bool:
        return bool(self.__config.get_value(INGESTER_LOG_UPLOAD_ERRORS))

    def get_should_raise_exception_on_failure(self) -> bool:
        return bool(self.__config.get_value(INGESTER_SHOULD_RAISE_EXCEPTION_ON_FAILURE))

    def get_payload_aggregator_timeout_seconds(self) -> int:
        return int(
            self.__config.get_value(INGESTION_PAYLOAD_AGGREGATOR_TIMEOUT_SECONDS)
        )


def add_definitions(config_builder: ConfigurationBuilder):
    # IngesterBase-related configurations
    config_builder.add(
        key=INGESTER_NUMBER_OF_WORKER_THREADS,
        default_value=10,
        description="Number of worker threads for async ingestion.",
    )
    config_builder.add(
        key=INGESTER_PAYLOAD_SIZE_BYTES_THRESHOLD,
        default_value=500 * 1024,  # Set default to 500KB
        description="""
        Aggregated payload threshold in bytes (when uncompressed). If the combined size of two or more payloads
        passed to the `send_object_for_ingestion()` or
        `send_tabular_data_for_ingestion()` methods is below this threshold, they
        will be bundled together, before being send to the ingestion plugin.
        Ingester make sure to send batches of data in approximately that size.
        Adjust size so that it is optimize for ingestion method. For example in Kafka 1KB may be better value.
        This is similar to INGESTION_PAYLOAD_AGGREGATOR_TIMEOUT_SECONDS and INGESTER_PAYLOADS_QUEUE_SIZE.
        """,
    )
    config_builder.add(
        key=INGESTION_PAYLOAD_AGGREGATOR_TIMEOUT_SECONDS,
        default_value=2,
        description="""
        Wait time in seconds, that a ingester thread is waiting for receiving a payload, before it continues with the next one.
        If payloads sit in the queue more than specified time they'd be bundled together and send to the ingestion plugin.
        This is similar to INGESTER_PAYLOAD_SIZE_BYTES_THRESHOLD and INGESTER_PAYLOADS_QUEUE_SIZE.
        """,
    )
    config_builder.add(
        key=INGESTER_PAYLOADS_QUEUE_SIZE,
        default_value=50,
        description="""
        If the number of payloads buffer reach this limit, they'd be bundled together and send to the ingestion plugin.
        This is similar to INGESTER_PAYLOAD_SIZE_BYTES_THRESHOLD and INGESTION_PAYLOAD_AGGREGATOR_TIMEOUT_SECONDS.
        """,
    )
    config_builder.add(
        key=INGESTER_OBJECTS_QUEUE_SIZE,
        default_value=10000,
        description="""
        The maximum number of payloads that can be stored for ingestion at any
        given time. When limit is reach send_*_for_ingestion method would effctively block unitl the queue is freed.
        """,
    )
    config_builder.add(
        key=INGESTER_LOG_UPLOAD_ERRORS,
        default_value=True,
        description="Specifies if errors should be logged.",
    )
    config_builder.add(
        key=INGESTER_SHOULD_RAISE_EXCEPTION_ON_FAILURE,
        default_value=True,
        description="When set to true, if there is an ingesting error, and we fail to ingest some data,"
        " ingester will raise an exception at the end - during finalize_job phase (plugin hook)."
        "By default this will cause the data job to fail (this behaviour can be overridden by plugins).",
    )



# Copyright 2021 VMware, Inc.
# SPDX-License-Identifier: Apache-2.0
import logging
from typing import Any
from typing import cast
from typing import Collection
from typing import List
from typing import Tuple
from typing import Union

from vdk.internal.builtin_plugins.connection.pep249.interfaces import PEP249Cursor


class ManagedCursor(PEP249Cursor):
    """
    PEP249 Cursor
    """

    def __init__(self, cursor: Any, log: logging.Logger = None) -> None:
        if not log:
            log = logging.getLogger(__name__)
        super().__init__(cursor, log)
        self._cur = cursor
        self._log = log

    def execute(
        self, operation: str, parameters: Union[List, Tuple] = None
    ) -> None:  # @UnusedVariable
        self._log.debug("Executing query:\n%s" % operation)
        try:
            if parameters:
                self._cur.execute(operation, parameters)
            else:
                self._cur.execute(operation)
            self._log.debug("Executing query SUCCEEDED.")
        except:
            self._log.debug("Executing query FAILED.")
            raise

    def fetchall(self) -> Collection[Collection[Any]]:
        self._log.debug("Fetching all results from query ...")
        try:
            res = self._cur.fetchall()
            self._log.debug("Fetching all results from query SUCCEEDED.")
            return cast(Collection[Collection[Any]], res)
        except:
            self._log.debug("Fetching all results from query FAILED.")
            raise

    def close(self) -> None:
        self._log.debug("Closing DB cursor ...")
        self._cur.close()
        self._log.debug("Closing DB cursor SUCCEEDED.")



# Copyright 2021 VMware, Inc.
# SPDX-License-Identifier: Apache-2.0
from __future__ import annotations

import logging
from collections import OrderedDict
from dataclasses import dataclass
from dataclasses import field
from typing import Any

from vdk.internal.core.errors import VdkConfigurationError

# Consider ConfigValue should be primitive type perhaps? and not just any object
ConfigValue = Any
ConfigKey = str

log = logging.getLogger(__name__)


def convert_value_to_type_of_default_type(
    key: ConfigKey, v: ConfigValue, default_value: ConfigValue
) -> ConfigValue:
    """
    Allows for configurations to be converted to the type of their default value.

    E.g. IMPALA_PORT is converted to int, because it's default value is int.
    """
    if default_value is not None:
        if type(default_value) == type(True) and type(v) != type(True):
            allowed_values = ["true", "false", "1", "0", "yes", "no", "y", "n"]
            if str(v).lower() not in allowed_values:
                msg = (
                    f"Provided configuration "
                    f'"{key}={v}" is invalid. Allowed values for {key} are {allowed_values}'
                )
                raise VdkConfigurationError(msg)
            v = str(v).lower() in ["true", "1", "yes", "y"]
        else:
            try:
                v = type(default_value)(v)  # cast to type of default_value:
            except ValueError:
                msg = (
                    f'Provided configuration "{key}={v}" is invalid. '
                    f'Cannot cast "{v}" to {str(type(default_value).__name__)}'
                )
                raise VdkConfigurationError(msg)
    return v


def _normalize_config_key(key: ConfigKey):
    return str(key).lower()


@dataclass(frozen=True)
class Configuration:
    """
    The configuration of the application. It's immutable.

    Use ConfigurationBuilder to build it.
    """

    __config_key_to_description: dict[ConfigKey, str]
    __config_key_to_value: dict[ConfigKey, ConfigValue]
    __config_key_to_default_value: dict[ConfigKey, ConfigValue] = field(
        default_factory=dict
    )

    def __getitem__(self, key: ConfigKey):
        key = _normalize_config_key(key)
        return self.get_value(key)

    def get_value(self, key: ConfigKey) -> ConfigValue:
        """
        Return configuration value associated with a given key or None if it does not exists.

        For required configuration see get_required_value

        :param key: the configuration key (e.g db_host, service_uri, etc.)
        :return: the value corresponding to the configuration key
        """
        key = _normalize_config_key(key)
        default_value = self.__config_key_to_default_value.get(key)
        value = self.__config_key_to_value.get(key, default_value)
        return value

    def get_required_value(self, key: ConfigKey) -> ConfigValue:
        """
        Return configuration value associated with a given key or throws VdkConfigurationError if it's missing.

        Use get_value if configuration is not required

        :param key: the configuration key (e.g db_host, service_uri, etc.)
        :return: the value corresponding to the configuration key
        :raises VdkConfigurationError
        """
        key = _normalize_config_key(key)
        value = self.get_value(key)
        if value is None:
            raise VdkConfigurationError(
                f"Required configuration {key} is missing."
                f"This will cause configuration error and some functionality may not work."
                f"Please provide correct configuration value. "
                f"See help for how to configure the tool"
            )
        return value

    def get_description(self, key: ConfigKey) -> str | None:
        """
        Get description of hte configuration with that key. It will be used to be printed in help.

        :param key: the config key
        :return: description
        """
        key = _normalize_config_key(key)
        return self.__config_key_to_description.get(key)

    def list_config_keys(self) -> list[ConfigKey]:
        """
        List all added (defined) config keys

        :return: list of key names.
        """
        return [k for k in self.__config_key_to_default_value.keys()]


@dataclass()
class ConfigurationBuilder:
    """
    Builder used to configure the app.

    Usually a plugin would define what variables it needs using add method.
    While other plugins will populate them using set_value method (e.g environment variable plugin. yaml file plugin, etc.).
    """

    __config_key_to_description: dict[ConfigKey, str]
    __config_key_to_value: dict[ConfigKey, ConfigValue]
    __config_key_to_default_value: dict[ConfigKey, ConfigValue]

    def __init__(self):
        self.__config_key_to_description = dict()
        self.__config_key_to_default_value = dict()
        self.__config_key_to_value = dict()

    def add(
        self,
        key: ConfigKey,
        default_value: ConfigValue,
        show_default_value=True,
        description=None,
    ) -> ConfigurationBuilder:
        """
        Add new configuration variable definition.

        :param key: The configuration key. If already exist variable will be updated.
        :param default_value: Default value for the configuration Can be None.
        :param show_default_value: default value will appear in help as well.
        The default value type will enforce the type of the option. Can be None - in this case the type would str.
        :param description: Set description if you want config variable to appear in command line help .
        It is strongly recommended to set description. If no description is set the config key will be hidden.
        TODO: in the future we should require description always and have separate hidden=True/False instead
        :return: self so it can be chained like builder.add(..).set_value(...)...
        """
        key = _normalize_config_key(key)
        self.__config_key_to_default_value[key] = default_value
        if description and show_default_value:
            self.__add_public(key, description, default_value)
        elif description:
            self.__add_public(key, description)
        return self

    def set_value(self, key: ConfigKey, value: ConfigValue) -> ConfigurationBuilder:
        """
        This will set or update the value for a configuration with a given key.

        It will try to cast it to the specified default type inferred from default value (set with #add method)
        :param key: the configuration key
        :param value: the configuration value.
        :return: self so it can be chained like builder.set_value(..).add(...)...
        """
        key = _normalize_config_key(key)
        default_value = self.__config_key_to_default_value.get(key)
        self.__config_key_to_value[key] = convert_value_to_type_of_default_type(
            key, value, default_value
        )
        return self

    def list_config_keys(self) -> list[ConfigKey]:
        """
        List all added (defined) config keys

        :return: list of key names.
        """
        return [k for k in self.__config_key_to_default_value.keys()]

    def __add_public(
        self, key: ConfigKey, description: str, default_value: ConfigValue = None
    ) -> None:
        if not isinstance(description, str):
            log.warning(
                f"Description for key {key} is not of type string. Converting to type string."
            )
            description = str(description)

        if default_value is not None:
            description += "\nDefault value is: '%s'." % default_value
        self.__config_key_to_description[key] = description

    def build(self) -> Configuration:
        """
        :return: immutable version of the Configuration
        """
        return Configuration(
            self.__config_key_to_description,
            self.__config_key_to_value,
            self.__config_key_to_default_value,
        )



# Copyright 2021 VMware, Inc.
# SPDX-License-Identifier: Apache-2.0
import inspect
from typing import Any


def class_fqname(py_object: Any) -> str:
    """
    :param py_object: Any python object
    :return: the full qualified name of the class (package.package.classname)
    """
    module = inspect.getmodule(py_object)
    if module is None:
        return py_object.__class__.__name__
    return f"{py_object.__class__.__module__}.{py_object.__class__.__name__}"



# Copyright 2021 VMware, Inc.
# SPDX-License-Identifier: Apache-2.0
"""
errors -- Exception handling implementations

errors is a module that handles all errors.
It defines classes and methods for handling exceptions, and ensuring that there are two types of exceptions:
 - infrastructure-related: should be handled by Platform team.
 - user code-related: should be handled by the owner of the DataJob.
"""
from __future__ import annotations

import logging
import re
import sys
import traceback
from collections import defaultdict
from collections import namedtuple
from enum import Enum
from logging import Logger
from types import TracebackType
from typing import Any
from typing import cast


class ResolvableBy(str, Enum):
    """
    Type of errors being thrown by VDK during execution of some command based on who is responsible for resolving/fixing them

    Those are:

    * PLATFORM_ERROR - for infrastructure errors that can and should be fixed by SRE Team, Platform team, operating the infrastructure and services.
    * USER_ERROR - Errors in user code/configuration, that should be fixed by the end user (or job owner) for example: supplied bad arguments, bug in user code.
    * CONFIG_ERROR - Errors in the configuration provided to VDK. Should be fixed by Platform if run in Platfrom infrastructure, or by end user, when run locally.
    """

    PLATFORM_ERROR = "Platform error"
    USER_ERROR = "User Error"
    CONFIG_ERROR = "Configuration error"


# TODO: instead of global variable set this in JobContext
#  (one benefit this way we can run even multiple jobs/templates in the same process)
# The key is 'blamee' (i.e. responsible for the error fixing) and the value is a list of corresponding ErrorMessages
BLAMEES: dict[str, list[Any]] = defaultdict(list)

# overide when running in kubernetes
CONFIGURATION_ERRORS_ARE_TO_BE_RESOLVED_BY = ResolvableBy.PLATFORM_ERROR

log = logging.getLogger(__name__)


class BaseVdkError(Exception):
    """For all errors risen from our "code" (vdk)

    There are two child branches in exception hierarchy:
     - Service errors
     - Domain errors
    """

    def __init__(self, error_message: ErrorMessage):
        super().__init__(str(error_message))


class PlatformServiceError(BaseVdkError):
    """
    Error caused by issue in Platform

    Service error are those errors which indicate for bad condition of the service. which prevent the normal
    completion of given request. User can do little more beyond retrying.
    """

    pass


class DomainError(BaseVdkError):
    """
    Domain errors are always function of the input and the current system state.

    In contrast to the service errors they occur always upon same conditions.
    In most of the cases the client can determine what is wrong and knows how to recover.
    """

    pass


class VdkConfigurationError(DomainError):
    """
    An exception in the configuration of the Data Job

    If job is executed on user (local execution) premises - this is considered #RESOLVED_BY_JOB_OWNER
    Otherwise If job is executed in "cloud" - this is considered #RESOLVED_BY_PLATFORM class of error
    """

    pass


class UserCodeError(DomainError):
    """Exception in user code"""

    pass


class ErrorMessage(
    namedtuple(
        "ErrorMessage", ["summary", "what", "why", "consequences", "countermeasures"]
    )
):
    """
    Standard format for Error messages in VDK. Use it when throwing exceptions or logging error level.
    """

    __slots__ = ()

    def _get_template(self, replace_with: str) -> str:
        return "{}r%15s : {}r%15s : {}r%15s : {}r%15s : {}".replace(
            "r", replace_with
        ) % (
            "what happened".upper(),
            "why it happened".upper(),
            "consequences".upper(),
            "countermeasures".upper(),
        )

    def _to_string(self, template: str) -> str:
        return template.format(
            self.summary, self.what, self.why, self.consequences, self.countermeasures
        )

    def __str__(self) -> str:
        return self._to_string(self._get_template("\n"))

    def to_html(self) -> str:
        """
        :return: html representaiton of the error.
        """
        return self._to_string(self._get_template("<br />"))


def get_blamee_overall() -> str | None:
    """
    Finds who is responsible for fixing the error/s.

    Returns:

    None - if there were no errors
    ResolvableBy.PLATFORM_ERROR - if during the run there were only Platfrom exception
    ResolvableBy.USER_ERROR - if during the run there was at least one job owner exceptions

    The reason it is set to ResolvableBy.USER_ERROR if there is at least one job owner is:
    VDK defaults the error to be resolved by Platform team  until it can determine for certain that it's an issue in the job code.
    There might be multiple components logging error (for example in exception propagation), and if at least one component says error is in the job code,
    then we set to be resolved by data job owner.

    """
    if len(BLAMEES) == 0:
        return None
    else:
        return (
            ResolvableBy.USER_ERROR
            if ResolvableBy.USER_ERROR in BLAMEES
            else ResolvableBy.PLATFORM_ERROR
        )


def get_blamee_overall_user_error() -> str:
    """
    Finds the first encountered error where the blamee overall is the owner.

    :return:
       Empty string if the owner is not the overall blamee
       An ErrorMessage instance
    """
    blamee = get_blamee_overall()
    if blamee is None or blamee == ResolvableBy.PLATFORM_ERROR:
        return ""
    blamee_errors = BLAMEES.get(blamee, [])
    return str(blamee_errors[0])


def get_error_type() -> str | None:
    """
    :return: "User" or "Platform"
    """
    blamee = get_blamee_overall()
    return (
        "User" if blamee == ResolvableBy.USER_ERROR else "Platform" if blamee else None
    )


def _build_message_for_end_user(
    to_be_fixed_by: ResolvableBy,
    what_happened: str,
    why_it_happened: str,
    consequences: str,
    countermeasures: str,
) -> ErrorMessage:
    error = ""
    if ResolvableBy.PLATFORM_ERROR == to_be_fixed_by:
        error = " Platform service error "
    elif ResolvableBy.USER_ERROR == to_be_fixed_by:
        error = "n error in data job code "
    elif ResolvableBy.CONFIG_ERROR == to_be_fixed_by:
        error = " configuration error "

    current_error_responsible_for_resolution = _error_type_to_actual_resolver(
        to_be_fixed_by
    )
    # statement to add the key in the dictionary (if not already there),
    # for get_blamee_overall()' to be calculated correctly
    BLAMEES[current_error_responsible_for_resolution]
    responsible_for_resolution = get_blamee_overall()

    msg = ErrorMessage(
        "A{} occurred. The error should be resolved by {}. Here are the details:".format(
            error, responsible_for_resolution
        ),
        what_happened.strip(),
        why_it_happened.strip(),
        consequences.strip(),
        countermeasures.strip(),
    )

    BLAMEES[current_error_responsible_for_resolution].append(msg)
    return msg


def get_caller_stacktrace() -> str:
    """
    :return: stacktrace excluding this method (hence caller stacktrace)
    """
    info = sys.exc_info()
    tb = cast(TracebackType, info[2])
    f = tb.tb_frame.f_back
    lst = ["Traceback (most recent call first):\n"]
    fstack = traceback.extract_stack(f)
    fstack.reverse()
    lst = lst + traceback.format_list(fstack)
    lines = ""
    for line in lst:
        lines = lines + line
    return lines


def _error_type_to_actual_resolver(to_be_fixed_by: ResolvableBy) -> str:
    if ResolvableBy.CONFIG_ERROR == to_be_fixed_by:
        return CONFIGURATION_ERRORS_ARE_TO_BE_RESOLVED_BY
    else:
        return to_be_fixed_by


def log_and_throw(
    to_be_fixed_by: ResolvableBy,
    log: Logger,
    what_happened: str,
    why_it_happened: str,
    consequences: str,
    countermeasures: str,
) -> None:
    """
    Log error message and then throw it to be handled up the stack.
    """
    msg = _build_message_for_end_user(
        to_be_fixed_by, what_happened, why_it_happened, consequences, countermeasures
    )

    try:
        if ResolvableBy.PLATFORM_ERROR == to_be_fixed_by:
            raise PlatformServiceError(msg)
        elif ResolvableBy.USER_ERROR == to_be_fixed_by:
            raise UserCodeError(msg)
        elif ResolvableBy.CONFIG_ERROR == to_be_fixed_by:
            raise VdkConfigurationError(msg)
        else:
            raise Exception(
                "BUG! Fix me!"
            )  # What type is the error that caused this and whom to blame Platform or Data Jobs Developer?
    except BaseVdkError as e:
        lines = get_caller_stacktrace()
        log.error(str(msg) + "\n" + lines)
        __set_error_is_logged(e)
        raise


def log_exception(
    to_be_fixed_by: ResolvableBy,
    log: Logger,
    what_happened: str,
    why_it_happened: str,
    consequences: str,
    countermeasures: str,
    exception: BaseException,
) -> None:
    """
    Log message only if it has not been logged already.
    Does not throw it again.
    """
    if __error_is_logged(exception):
        return
    msg = _build_message_for_end_user(
        to_be_fixed_by, what_happened, why_it_happened, consequences, countermeasures
    )
    __set_error_is_logged(exception)
    log.exception(msg)


def wrap_exception_if_not_already(
    to_be_fixed_by: ResolvableBy, msg: ErrorMessage, exception: BaseException
):
    if isinstance(exception, BaseVdkError):
        # already wrapped
        return exception

    # TODO: how to assign cause (new_ex from old_ex) ?
    if ResolvableBy.PLATFORM_ERROR == to_be_fixed_by:
        return PlatformServiceError(msg)
    elif ResolvableBy.USER_ERROR == to_be_fixed_by:
        return UserCodeError(msg)
    elif ResolvableBy.CONFIG_ERROR == to_be_fixed_by:
        return VdkConfigurationError(msg)
    else:
        log.warning(
            "Unknown to_be_fixed_by type. "
            "This seems like a bug. We cannot wrap exception and return original one "
        )
        return exception


def log_and_rethrow(
    to_be_fixed_by: ResolvableBy,
    log: Logger,
    what_happened: str,
    why_it_happened: str,
    consequences: str,
    countermeasures: str,
    exception: BaseException,
    wrap_in_vdk_error=False,
) -> None:
    """
    Log message only if it has not been logged already. And throws it again. Use it to handle coming exceptions.
    :param to_be_fixed_by:
    :param log:
    :param what_happened: same as ErrorMessage#what
    :param why_it_happened: same as ErrorMessage#why
    :param consequences: same as ErrorMessage#consequences
    :param countermeasures: same as ErrorMessage#countermeasures
    :param exception: the exception message to re-throw
    :param wrap_in_vdk_error: If the exception is not wrapped by BaseVdkError
            it will wrap it in corresponding BaseVdkError exception based on to_be_fixed_by parameter
    """

    msg = _build_message_for_end_user(
        to_be_fixed_by, what_happened, why_it_happened, consequences, countermeasures
    )
    to_be_raised_exception = exception
    if wrap_in_vdk_error:
        to_be_raised_exception = wrap_exception_if_not_already(
            to_be_fixed_by, msg, exception
        )

    if not __error_is_logged(exception):
        log.exception(msg)
        __set_error_is_logged(exception)

    if wrap_in_vdk_error:
        raise to_be_raised_exception from exception
    else:
        raise exception


def __error_is_logged(exception: BaseException) -> bool:
    """Check if exception has custom added attribute is_logged"""
    return hasattr(exception, "is_logged")


def __set_error_is_logged(exception: BaseException):
    setattr(exception, "is_logged", True)


def find_whom_to_blame_from_exception(exception: Exception) -> ResolvableBy:
    """
    Tries to determine if it's user or platform error
    """

    if issubclass(type(exception), UserCodeError):
        return ResolvableBy.USER_ERROR
    if issubclass(type(exception), VdkConfigurationError):
        return (
            ResolvableBy.CONFIG_ERROR
        )  # TODO find out if this is a local or platform deployment and fix this line.
    if issubclass(type(exception), PlatformServiceError):
        return ResolvableBy.PLATFORM_ERROR

    return ResolvableBy.PLATFORM_ERROR


MSG_CONSEQUENCE_DELEGATING_TO_CALLER__LIKELY_EXECUTION_FAILURE = (
    "I'm rethrowing this exception to my caller to process."
    "Most likely this will result in failure of current Data Job."
)
MSG_CONSEQUENCE_TERMINATING_APP = (
    "The provided Data Job will not be executed. Terminating application."
)
MSG_COUNTERMEASURE_FIX_PARENT_EXCEPTION = (
    "See contents of the exception and fix the problem that causes it."
)


def get_exception_message(exception: Exception) -> str:
    """Returns the message part of an exception as string"""
    return str(exception).strip()


class CustomMessageExceptionDecorator:
    """
    Provides custom message for an exception.

    Needed in order for the end-user to be better informed for the exceptional condition.
    """

    def __init__(self, exception: Exception) -> None:
        self._exception = exception

    def get_custom_message(self) -> str | None:
        """
        Attach custom message for end-user to exception.

        :return: decorated exception message.
        """
        return self._decorate_exception_message(self._exception)

    @staticmethod
    def _decorate_exception_message(exception: Exception) -> str | None:
        if exception_matches(exception, "OSError", ".*File too large.*"):
            message = (
                "File size limit  for Data Job has been exceeded."
                "Optimize the disk utilization of your Data Job."
                "For local runs this limit can be changed by 'export VDK_RESOURCE_LIMIT_DISK_MB=<new-value>'."
                "For cloud run you can fill in a Service Request (http://go/resource-limits) for the limit to be changed."
                "This limit does not apply to macOS and Windows execution environments."
            )
            return f"An exception occurred, exception message was: {message}"

        if exception_matches(exception, "RuntimeError", ".*can't start new thread.*"):
            message = (
                "Unable to start new thread. Optimize thread usage of your Data Job."
            )
            return f"An exception occurred, exception message was: {message}"
        return None


def MSG_WHY_FROM_EXCEPTION(exception: Exception) -> str:
    """
    Try to figure what is the reason for the failure (why) from the exception and return as a reason.
    """
    custom_message = CustomMessageExceptionDecorator(exception).get_custom_message()
    if custom_message:
        return custom_message
    else:
        return "An exception occurred, exception message was: {}".format(
            get_exception_message(exception)
        )


def exception_matches(
    e: Exception, classname_with_package: str, exception_message_matcher_regex: str
) -> bool:
    """
    Tries to see if exception matches given regex expression. Use to detect what type of error is.

    :param e: BaseException: exception whose class and message are to be checked
    :param classname_with_package: full class name, e.g. 'impala.error.HiveServer2Error'
    :param exception_message_matcher_regex: A regex expression that is run against exception message.
    """
    typeMatcher = re.compile(classname_with_package)
    msgMatcher = re.compile(pattern=exception_message_matcher_regex, flags=re.DOTALL)
    t = type(e)  # "<class 'impala.error.HiveServer2Error'>"
    classname = str(t).split("'")[1]
    match = typeMatcher.match(classname)
    if None is match:
        return False
    grp = match.group()
    if not (grp == classname):
        return False

    msg = get_exception_message(e)
    match = msgMatcher.match(msg)
    if None is match:
        return False
    grp = match.group()
    return grp == msg


def clear_intermediate_errors() -> None:
    """
    Clear so far recorded records - those are considered intermediate and resolved.

    For example after successful completion of a step.
    """
    BLAMEES.clear()



# Copyright 2021 VMware, Inc.
# SPDX-License-Identifier: Apache-2.0
import logging
import time
import uuid
from datetime import datetime
from typing import List

import click
from vdk.api.plugin.core_hook_spec import JobRunHookSpecs
from vdk.api.plugin.hook_markers import hookimpl
from vdk.internal import vdk_build_info
from vdk.internal.builtin_plugins.config import vdk_config
from vdk.internal.builtin_plugins.config.config_help import ConfigHelpPlugin
from vdk.internal.builtin_plugins.config.log_config import LoggingPlugin
from vdk.internal.builtin_plugins.config.vdk_config import CoreConfigDefinitionPlugin
from vdk.internal.builtin_plugins.config.vdk_config import EnvironmentVarsConfigPlugin
from vdk.internal.builtin_plugins.config.vdk_config import JobConfigIniPlugin
from vdk.internal.builtin_plugins.debug.debug import DebugPlugins
from vdk.internal.builtin_plugins.ingestion.ingester_configuration_plugin import (
    IngesterConfigurationPlugin,
)
from vdk.internal.builtin_plugins.job_properties.properties_api_plugin import (
    PropertiesApiPlugin,
)
from vdk.internal.builtin_plugins.notification.notification import NotificationPlugin
from vdk.internal.builtin_plugins.termination_message.writer import (
    TerminationMessageWriterPlugin,
)
from vdk.internal.builtin_plugins.version.new_version_check_plugin import (
    NewVersionCheckPlugin,
)
from vdk.internal.core.config import ConfigurationBuilder
from vdk.internal.core.context import CoreContext
from vdk.internal.core.statestore import CommonStoreKeys
from vdk.internal.plugin.plugin import PluginRegistry

log = logging.getLogger(__name__)


class RuntimeStateInitializePlugin:
    """
    Initialize in the StateStore common runtime attributes
    """

    @hookimpl(tryfirst=True)
    def vdk_initialize(self, context: CoreContext) -> None:
        """
        Setup common state attributes of the app (CommonStoreKeys).

        Configuration can override an attempt/op/execution id.

        Attempt id and Execution id are extracted from env variables.
        If not present they are auto generated.

        If auto generated:
        * Attempt id format is uuid + timestamp + five random chars.
        * Execution id is attempt id minus last 6 chars.
        * Op id is equal to execution id.
        """
        op_id = context.configuration.get_value(vdk_config.OP_ID)
        execution_id = context.configuration.get_value(vdk_config.EXECUTION_ID)
        attempt_id = context.configuration.get_value(vdk_config.ATTEMPT_ID)

        if not attempt_id:
            # Generate attempt id if execution id not present.
            if not execution_id:
                # UUID is 36 chars
                attempt_id = f"{str(uuid.uuid4())}-{str(int(time.time()))}-{str(uuid.uuid4())[:5]}"
            # Use execution id to generate attempt id.
            else:
                attempt_id = f"{execution_id}-{str(uuid.uuid4())[:5]}"
            # If env is cloud we must print a warning.
            if context.configuration.get_value(vdk_config.LOG_CONFIG) == "CLOUD":
                log.warning(
                    f"Attempt ID not found in env or configuration. Using auto generated attempt_id: {attempt_id}"
                )

        if not execution_id:
            # Delete everything after the last '-' character in the attempt id (inclusive)
            execution_id = "-".join(attempt_id.split("-")[:-1])

        if not op_id:
            op_id = execution_id

        log.info(
            f"Setting: OP_ID: {op_id}, ATTEMPT_ID: {attempt_id}, EXECUTION_ID: {execution_id}"
        )
        context.state.set(CommonStoreKeys.OP_ID, op_id)
        context.state.set(CommonStoreKeys.EXECUTION_ID, execution_id)
        context.state.set(CommonStoreKeys.ATTEMPT_ID, attempt_id)

        context.state.set(CommonStoreKeys.VDK_VERSION, vdk_build_info.RELEASE_VERSION)
        context.state.set(CommonStoreKeys.START_TIME, datetime.utcnow())


@hookimpl
def vdk_start(plugin_registry: PluginRegistry, command_line_args: List) -> None:
    """
    Load all default (builtin) vdk plugins
    """
    log.info("Load default (builtin) vdk plugins.")
    plugin_registry.load_plugin_with_hooks_impl(DebugPlugins())
    plugin_registry.load_plugin_with_hooks_impl(LoggingPlugin())
    plugin_registry.load_plugin_with_hooks_impl(ConfigHelpPlugin())
    plugin_registry.load_plugin_with_hooks_impl(CoreConfigDefinitionPlugin())
    plugin_registry.load_plugin_with_hooks_impl(EnvironmentVarsConfigPlugin())
    plugin_registry.load_plugin_with_hooks_impl(RuntimeStateInitializePlugin())
    plugin_registry.load_plugin_with_hooks_impl(NewVersionCheckPlugin())
    plugin_registry.load_plugin_with_hooks_impl(NotificationPlugin())
    plugin_registry.load_plugin_with_hooks_impl(IngesterConfigurationPlugin())
    plugin_registry.load_plugin_with_hooks_impl(PropertiesApiPlugin())
    # TODO: should be in run package only
    plugin_registry.add_hook_specs(JobRunHookSpecs)
    plugin_registry.load_plugin_with_hooks_impl(JobConfigIniPlugin())
    plugin_registry.load_plugin_with_hooks_impl(TerminationMessageWriterPlugin())


@hookimpl(tryfirst=True)
def vdk_command_line(root_command: click.Group) -> None:
    """
    Register commands
    """
    from vdk.internal.builtin_plugins.run.cli_run import run
    from vdk.internal.builtin_plugins.version.version import version

    root_command.add_command(run)
    root_command.add_command(version)




RELEASE_VERSION='0.0.393010107'
BUILD_DATE='Thu Oct 21 17:58:23 UTC 2021'
BUILD_MACHINE_INFO='Linux runner-mvj-ljo-project-28359933-concurrent-0xnhr6 5.4.129-62.227.amzn2.x86_64 #1 SMP Wed Jul 7 00:08:43 UTC 2021 x86_64 GNU/Linux'
GITLAB_CI_JOB_ID='1703331830'
GIT_COMMIT_SHA='e294b016729996a317363efc069024036b876312'
GIT_BRANCH='main'



# Copyright 2021 VMware, Inc.
# SPDX-License-Identifier: Apache-2.0
import logging
import pathlib
from typing import Dict

from vdk.api.job_input import ITemplate
from vdk.api.plugin.plugin_input import ITemplateRegistry
from vdk.internal.builtin_plugins.run.data_job import DataJobFactory
from vdk.internal.builtin_plugins.run.execution_results import ExecutionResult
from vdk.internal.core import errors
from vdk.internal.core.context import CoreContext

log = logging.getLogger(__name__)


class TemplatesImpl(ITemplateRegistry, ITemplate):
    def __init__(
        self,
        job_name: str,
        core_context: CoreContext,
        datajob_factory: DataJobFactory = None,
    ):
        self._job_name = job_name
        self._core_context = core_context
        self._registered_templates: Dict[str, pathlib.Path] = {}
        self._datajob_factory = (
            DataJobFactory() if datajob_factory is None else datajob_factory
        )

    def add_template(self, name: str, template_directory: pathlib.Path):
        if (
            name in self._registered_templates
            and self._registered_templates[name] != template_directory
        ):
            log.warning(
                f"Template with name {name} has been registered with directory {self._registered_templates[name]}."
                f"We will overwrite it with new directory {template_directory} now."
            )
        self._registered_templates[name] = template_directory

    def execute_template(self, name: str, template_args: Dict) -> ExecutionResult:
        log.debug(f"Execute template {name} {template_args}")
        template_directory = self.get_template_directory(name)
        template_job = self._datajob_factory.new_datajob(
            template_directory, self._core_context, name=self._job_name
        )
        result = template_job.run(template_args)
        if result.is_failed():
            if result.get_exception():
                raise result.get_exception()
            else:
                errors.log_and_throw(
                    errors.ResolvableBy.PLATFORM_ERROR,
                    logging.getLogger(__name__),
                    f"Template `{name}` failed.",
                    f"No exception is reported. This is not expected. "
                    f"Execution steps of templates were {result.steps_list}",
                    "We will raise an exception now. Likely the job will fail.",
                    f"Check out the error and fix the template invocation "
                    f"or fix the template by installing the correct plugin."
                    f" Or open an issue on the support team or "
                    f"Versatile Data Kit or the plugin provider of the template",
                )
        return result

    def get_template_directory(self, name: str) -> pathlib.Path:
        if name in self._registered_templates:
            return self._registered_templates[name]
        else:
            errors.log_and_throw(
                to_be_fixed_by=errors.ResolvableBy.USER_ERROR,
                log=logging.getLogger(__file__),
                what_happened=f"No registered template with name: {name}.",
                why_it_happened="Template with that name has not been registered",
                consequences=errors.MSG_CONSEQUENCE_DELEGATING_TO_CALLER__LIKELY_EXECUTION_FAILURE,
                countermeasures="Make sure you have not misspelled the name of the template "
                "or the plugin(s) providing the template is installed. "
                f"Current list of templated is: {list(self._registered_templates.keys())}",
            )



# Copyright 2021 VMware, Inc.
# SPDX-License-Identifier: Apache-2.0
from vdk.internal.core.config import Configuration
from vdk.internal.core.config import ConfigurationBuilder

PROPERTIES_DEFAULT_TYPE = "PROPERTIES_DEFAULT_TYPE"


class PropertiesConfiguration:
    def __init__(self, config: Configuration):
        self.__config = config

    def get_properties_default_type(self) -> str:
        return self.__config.get_value(PROPERTIES_DEFAULT_TYPE)


def add_definitions(config_builder: ConfigurationBuilder):
    config_builder.add(
        key=PROPERTIES_DEFAULT_TYPE,
        default_value=None,
        description="Set the default properties type to use. "
        "Plugins can register different properties types. "
        "This option controls which is in use"
        "It can be left empty, in which case "
        "if there is only one type registered it will use it."
        "Or it will use one register with type 'default' ",
    )
